#include <vector>
#include <stdlib.h> /* rand() */
#include <sys/time.h>

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/variate_generator.hpp>

#define BISQBOT_INTERNAL
#include "bisqbot.hh"
static const unsigned MAX_FRAMES = 240;

extern "C" unsigned char RAM[0x800];
extern "C" unsigned char PALRAM[0x20];

static unsigned NextChange=0;
static unsigned NextJump  =0;

static boost::mt19937 rng;

static class BisqBotState: public BisqBotStateBase
{
public:
    virtual ~BisqBotState() { }
    
    BisqBotState()
    {
        Reset();
    }
    void Reset()
    {
        PrevMarioX = 0;
        NFrames    = 0;
        MarioMovement = 0;
        Pit = 0;
        Entry.Clear();
        Damage.Clear();
        
        JumpRemain = 0;
        MarioY = 0;
        Active = 0;
    }
    
    void operator= (const BisqBotState& b)
    {
        PrevMarioX    = b.PrevMarioX;
        NFrames       = b.NFrames;
        MarioMovement = b.MarioMovement;
        Pit = b.Pit;
        Entry = b.Entry;
        Damage = b.Damage;
        JumpRemain = b.JumpRemain;
        MarioY = b.MarioY;
        Active = b.Active;
    }

public:
    unsigned PrevMarioX, NFrames;
    int MarioMovement;
    int Pit;
    BisqBotVar Entry, Damage;
    int MarioY;
    int Active;

    unsigned JumpRemain;
} state;

const BisqBotStateBase* BisqBotNewSaveState()
{
    return new BisqBotState(state);
}

void BisqBotLoadState(const BisqBotStateBase* b)
{
    state = *(BisqBotState*)b;
}

extern "C" {
void FCEUI_GetIVectors(unsigned short *reset, unsigned short *irq, unsigned short *nmi);
unsigned HScroll,VScroll;
}

void BisqBotMonitorCPU
  (unsigned short PC,
   unsigned char A,
   unsigned char X,
   unsigned char Y)
{
/*
    static bool InSoundCode=false;
    static bool InWaitFrame=false;
    static bool InNMI=false;
    if(PC == 0xD551) { InSoundCode=true; return; }
    if(PC == 0xD554) { InSoundCode=false; return; }
    if(InSoundCode) return;
    if(PC == 0xC01D) { InWaitFrame=true; return; }
    if(PC == 0xC059) { InWaitFrame=false; return; }
    if(InWaitFrame) return;
    if(PC == 0xD4A8) { InNMI=true; return; }
    if(PC == 0xD582) { InNMI=false; return; }
    if(InNMI) return;
    printf("PC[%04X]", PC);
*/
    static unsigned short irq=0;
    if(!irq) { unsigned short res, nmi; FCEUI_GetIVectors(&res,&irq,&nmi); }
    
    if(PC == irq)
    {
        static int LastX=-1,LastY=-1, frame=0;
        extern unsigned HScroll, VScroll;
        int NowX=HScroll, NowY=VScroll;
        
        if(NowX != LastX || NowY != LastY)
        {
            fprintf(stderr, "Frame %u, X:%3d Y:%3d", frame, NowX,NowY);
            for(unsigned a=0; a<0x20; ++a) fprintf(stderr, " %02X", PALRAM[a]);
            fprintf(stderr, "\n");
            
            LastX=NowX;
            LastY=NowY;
        }
        ++frame;
    }

#if 0 /* SMB */
    /* GAME MONITORING */
    /* A, X, Y, PC correspond to the values of the CPU registers. */
    if(PC == 0x8082)
    {
        unsigned MarioXPos = RAM[0x86];
        unsigned MarioYPos = RAM[0xCE];
        unsigned ScrollBegin = RAM[0x73F];
        
        state.MarioY = MarioYPos;
        
        unsigned MarioScreenX = (unsigned char)(MarioXPos-ScrollBegin+256);
        
        unsigned MarioX = MarioXPos;
        if(state.NFrames++ > 0)
        {
            //fprintf(stderr, "MarioX = %u\n", MarioX);
            state.MarioMovement += (signed char)(MarioX - state.PrevMarioX);
        }
        state.PrevMarioX = MarioX;
    }
    
    if(PC == 0xB04C) // magic_jT9+2
    {
        //fprintf(stderr, "JT9 with A=%02X\n", A);
        switch(A)
        {
            case 1: // auto climb
            case 2: // enter horiz pipe
            case 3: // enter vert pipe
            case 4: // slide flagpole down
            case 5: // victory 
            case 9: // eat mushroom
            case 12: // eat flower
                state.Entry.Set(A);
                break;
            case 10:// eat damage
            case 11:// eat death
                state.Damage.Set(A);
                fprintf(stderr, "Pit (a=%u)\n", A);
                state.Pit=1;
                break;
            case 0: // regen level
            case 7: // start level
            {
                state.Pit    = 0;
                state.MarioY = 128;
                break;
            }
            case 8: // normal level activity
            {
                state.Active = 1;
                
                int MarioYPos = state.MarioY;
                
                if(!state.Pit && MarioYPos >= 0xF0 && MarioYPos <= 0xFF)
                {
                    state.Pit = -1; // ignore, it's screen top
                }
                else if(state.Pit < 0 && MarioYPos < 0x60)
                    state.Pit = 0;
                else if(!state.Pit && MarioYPos > 0xC0 && MarioYPos < 0xFF)
                {
                    state.Pit=1;
                    fprintf(stderr, "Pit (y=%u)\n", MarioYPos);
                }
                break;
            }
        }
    }
    if(PC == 0xCFEC)
    {
        state.Entry.Set(15); // Killed koopa
    }
#endif
#if 1 /* Rockman3 */
    /* GAME MONITORING */
    /* A, X, Y, PC correspond to the values of the CPU registers. */
    if(PC == 0xC000)
    {
        unsigned MarioXPos = RAM[0x360];
        unsigned MarioX = MarioXPos;
        if(state.NFrames++ > 0)
        {
            //fprintf(stderr, "MarioX = %u\n", MarioX);
            state.MarioMovement += (signed char)(MarioX - state.PrevMarioX);
        }
        state.PrevMarioX = MarioX;
        state.Active=1;
    }
#endif
}

void BisqBotReset()
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    rng.seed(tv.tv_usec + tv.tv_sec*1000000ULL);
    
    state.Reset();
}

template<unsigned mod>
static inline unsigned GetRnd()
{
    static boost::uniform_int<int> g(0, mod-1);
    return g(rng);
}
static inline unsigned GetRnd(unsigned mod)
{
    return boost::uniform_int<int> (0,mod-1) (rng);
}

unsigned char BisqBotInput(unsigned FrameNo)
{
    if(state.Entry.Count()) return 0;
    
    //if(FrameNo >= 4) return K_A + K_B + K_R;
    
    static const unsigned char InputTable[] =
    {
        K_R,
        K_R,
        K_R,
        K_R,/*
        K_R,
        K_R,
        K_R,
        K_R,
        K_R,*/
        K_L,
        0
    };
    static unsigned LastInput = 0;
    static unsigned A_State   = 0;
    static unsigned B_State   = K_B;
    unsigned char Input = 0x00;
    
    if(!NextChange) 
    {
        LastInput = InputTable[ GetRnd<sizeof(InputTable)>() ];
        NextChange = GetRnd<10>();
    }
    else --NextChange;
    
    /*B_State = K_B;*/
    /*
    if(B_State)
        { if(GetRnd<100>() < 1) B_State=0; }
    else
        { if(GetRnd<100>() < 95) B_State=K_B; }
    */
    if(GetRnd<30>() == 0) B_State = K_B; else B_State = 0;
    if(GetRnd<20>() == 0) Input |= K_U;
    if(GetRnd<20>() == 0) Input |= K_D;

    /*
    if(!GetRnd<20>())
    {
        Input |= K_D;
        if(GetRnd<5>()) Input &= ~(K_R | K_L);
    }
    */
    
    if(!state.JumpRemain)
    {
        if(!NextJump)
        {
            state.JumpRemain = GetRnd(1+GetRnd<6>());
            NextJump   = GetRnd<10>() + 1;
        } else --NextJump;
    }
    
    Input |= LastInput;
    if(state.JumpRemain) Input |= K_A;
    Input |= B_State;
    
    return Input;
}

unsigned long long BisqBotEvaluate(unsigned FrameNo)
{
    //fprintf(stderr, "Joypad=%02X\n", RAM[0x6FC]);
    /*
    fprintf(stderr, "Evaluate %u, movement=%d, %u frames, nextchange=%u\n",
        FrameNo,
        state.MarioMovement,
        state.NFrames,
        NextChange);
    */
    
    if(!state.Active) return 912*1000000ULL;
    state.Active = 0;
    
    if(state.Pit > 0) return 0; // dead
    
    unsigned score1 = 0;
/*
    for(unsigned a=0; a<6; ++a)
        score1 = score1*10 + RAM[0x7DD+a];
    //fprintf(stderr, "score1: %u\n", score1);
*/
    
    unsigned movescore = state.MarioMovement + 512;
    unsigned goalscore = state.Entry.Count() * 2048;
    return (movescore + goalscore) * 1000000ULL + score1;
}

enum BisqBotInstruction BisqBotTick(unsigned FrameNo)
{
    if(state.Pit > 0) return BisqBotRestart;
    
    if(FrameNo > MAX_FRAMES) return BisqBotRestart;
    
    //fprintf(stderr, "%u: %d\n", FrameNo, MarioMovement);
    return BisqBotContinue;
}
