#include <vector>
#include <stdlib.h> /* rand() */
#include <time.h>

#define BISQBOT_INTERNAL
#include "bisqbot.hh"
static const unsigned MAX_FRAMES = 238;

static BisqBotVar Bonus, Hurt, Item, Kill, BossHit;
static BisqBotVar PauseActivated;
static unsigned LastTick;

extern "C" unsigned char RAM[0x800];
    
void BisqBotMonitorCPU(
   unsigned short PC,
   unsigned char A,
   unsigned char X,
   unsigned char Y)
{
    /* GAME MONITORING */
    /* A, X, Y, PC correspond to the values of the CPU registers. */
    
    static const unsigned short  FrameUpd   = 0xD57D;
    static const unsigned short  RandomFunc = 0xC5AB;
    static const unsigned short  CreateDrop = 0xBF25;
    static const unsigned short  HurtFunc   = 0xA244;
    static const unsigned short  GotItem    = 0xC833;
    static const unsigned short  KillFunc   = 0xC219;
    static const unsigned short  EnemyKill  = 0xBEEC;
    static const unsigned short  BossDamaged= 0xC90F;
    static const unsigned short  BossKilled = 0xC916;
    static const unsigned short SelectPushed= 0x91D1;
    static const unsigned short PauseRelease= 0x91F5;
    
    unsigned char GameFrame = RAM[0x23];
    unsigned char Bank      = RAM[0x42];
    
    #define TestPC(rule) \
        if(rule) fprintf(stderr, "[%u,%02X,%02X]Match %s (%04X,%02X,%02X,%02X)\n", \
            LastTick, GameFrame, Bank, \
            #rule, \
            PC,A,X,Y); \
        if(rule)
    
    TestPC(PC == CreateDrop)  Bonus.Set(Y);
    TestPC(PC == HurtFunc)    Hurt.Set(1);
    TestPC(PC == KillFunc)    Hurt.Set(2);
    TestPC(PC == GotItem)     Item.Set(A);
    TestPC(PC == EnemyKill)   Kill.Set(A);
    TestPC(PC == BossDamaged) BossHit.Set(1);
    TestPC(PC == BossKilled)  BossHit.Set(2);
    TestPC(PC == SelectPushed && Bank == 0x05 && A == 0x22)
    {
        PauseActivated.Set(1);
    }
    TestPC(PC == PauseRelease && Bank == 0x05 && A == RAM[0x95])
    {
    }
}

static bool PauseActive = false;
static bool WantPause = false;
static int BFrame=0, UnPause=0;

void BisqBotReset()
{
    Bonus.Clear();
    Hurt.Clear();
    Item.Clear();
    Kill.Clear();
    
    BossHit.Clear();
    PauseActivated.Clear();
    WantPause   = false;
    PauseActive = false;
    UnPause=0;
    
    BFrame = 8 + rand()%20;
    fprintf(stderr, "restart\n");
}

unsigned char BisqBotInput(unsigned FrameNo)
{
    static const unsigned char InputTable[] =
    {
        0x00, 0x01,
        0x40, 0x41,
        0x80, 0x81,
        0x01, 0x41, 0x81,
        0x01, 0x41, 0x81,
        0x01, 0x41, 0x81
    };
    static unsigned LastInput = 0x00;
    unsigned char Input = 0x00;
    //if(!(rand()%20)) Input |= 0x02;
    if(FrameNo==BFrame) Input |= 0x02;
    if(!(rand()%10)) LastInput = InputTable[rand()%sizeof(InputTable)];

    if(FrameNo < BFrame) LastInput = (LastInput&0x3F) | 0x80;
    
    Input |= LastInput;
    
    if(WantPause)
    {
        Input |= 0x04; //select button
    }
    else if(PauseActive)
    {
        if(FrameNo == UnPause)
        {
            Input |= 0x04;
            PauseActive=false;
            UnPause=0;
        }
    }
    
    return Input;
}

enum BisqBotInstruction BisqBotTick(unsigned FrameNo)
{
LastTick=FrameNo;
    if(BossHit.Has(1) && !BossHit.Has(2))
    {
        BossHit.Clear();
        WantPause = true;
    }
    if(PauseActivated.Count() > 0)
    {
        PauseActivated.Clear();
        if(!PauseActive)
        {
            WantPause=false;
            PauseActive=true;
            UnPause = FrameNo + 7+rand()%6;
            fprintf(stderr, "UnPause = %u\n", UnPause-FrameNo);
            srand(time(NULL));
        }
    }
    
    if(/*Hurt.Count()
    || Item.Count()
    || !Bonus.LimitedTo(0x3B, 0x3F)
    || */FrameNo > MAX_FRAMES)
    {
        return BisqBotRestart;
    }
    
    if(BossHit.Has(2))
    //if(Bonus.Has(0x3F, 1))
    {
        return BisqBotAccept;
    }
    return BisqBotContinue;
}
