#include "sdl.h"
#include "throttle.h"

static uint64 Lasttime, Nexttime;
static long double desired_frametime;
static double fps_scale = 1.0;
static int InFrame;

/* LOGMUL = exp(log(2) / 3) */
#define LOGMUL 1.259921049894873

void RefreshThrottleFPS(void)
{
  uint64 fps = FCEUI_GetDesiredFPS(); // Do >> 24 to get in Hz
  desired_frametime = 16777216.0l / (fps * fps_scale);

  Lasttime=0;
  Nexttime=0;
  InFrame=0;
}

int SpeedThrottle(void)
{
  if(fps_scale >= 32)
  {
    return 0;
  }
  uint64 time_left;
  uint64 cur_time;
  
  if(!Lasttime)
    Lasttime = SDL_GetTicks();
  
  if(!InFrame)
  {
    InFrame = 1;
    Nexttime = Lasttime + desired_frametime * 1000;
  }
  
  cur_time  = SDL_GetTicks();
  if(cur_time >= Nexttime)
    time_left = 0;
  else
    time_left = Nexttime - cur_time;
  
  if(time_left > 50)
    time_left = 50;
  else
    InFrame = 0;
  
  /*fprintf(stderr, "attempting to sleep %Ld ms, frame complete=%s\n",
    time_left, InFrame?"no":"yes");*/
  SDL_Delay(time_left);
  
  if(!InFrame)
  {
    Lasttime = SDL_GetTicks();
    return 0;
  }
  return 1;
}

void IncreaseEmulationSpeed(void)
{
  fps_scale *= LOGMUL;
  
  if(fps_scale > 32) fps_scale = 32;

  RefreshThrottleFPS();

  FCEU_DispMessage("emulation speed %.1f%%", fps_scale*100.0);
}

void DecreaseEmulationSpeed(void)
{
  fps_scale /= LOGMUL;
  if(fps_scale < 0.015625)
    fps_scale = 0.015625;

  RefreshThrottleFPS();

  FCEU_DispMessage("emulation speed %.1f%%", fps_scale*100.0);
}

void FCEUD_SetEmulationSpeed(int cmd)
{
    switch(cmd)
    {
    case EMUSPEED_SLOWEST:  fps_scale=0.02;  break;
    case EMUSPEED_SLOWER:   DecreaseEmulationSpeed(); break;
    case EMUSPEED_NORMAL:   fps_scale=1.0; break;
    case EMUSPEED_FASTER:   IncreaseEmulationSpeed(); break;
    case EMUSPEED_FASTEST:  fps_scale=10.0;break;
    default:
        return;
    }

    RefreshThrottleFPS();

    FCEU_DispMessage("emulation speed %.1f%%", fps_scale*100.0);
}  
